%{
/* Cherokee
 *
 * Authors:
 *      Alvaro Lopez Ortega <alvaro@alobbs.com>
 *
 * Copyright (C) 2001-2003 Alvaro Lopez Ortega
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <pwd.h>
#include <grp.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>

#include "mime.h"
#include "common.h"
#include "server.h"
#include "table.h"
#include "virtual_server.h"
#include "plugin_table_entry.h"
#include "encoder.h"
#include "logger_table.h"


/* Define the parameter name of the yyparse() argument
 */
#define YYPARSE_PARAM server

%}


%union {
	   int   number;
	   char *string;
	   void *ptr;

	   struct {
			 char *name;
			 void *ptr;
	   } name_ptr;

	   void *list;
}


%{
/* What is the right way to import this prototipe?
 */
extern int yylex(void);


extern char *yytext;
extern int   yylineno;

cherokee_virtual_server_t      *current_virtual_server     = NULL;
cherokee_plugin_table_entry_t  *current_plugin_table_entry = NULL;
cherokee_encoder_table_entry_t *current_encoder_entry      = NULL;
list_t                         *current_list               = NULL;

struct {
	   char                          *handler_name;
	   cherokee_handler_t            *handler;
	   cherokee_plugin_table_entry_t *entry;
	   cherokee_virtual_server_t     *vserver;
	   char                          *document_root;
	   char                          *directory_name;
} directory_content_tmp;

typedef struct {
	   void *next;
	   char *string;
} linked_list_t;


#define auto_virtual_server ((current_virtual_server) ? current_virtual_server : SRV(server)->vserver_default)


char *
make_finish_with_slash (char *string, int *len)
{
	   char *new;

	   if (string[(*len)-1] == '/') {
			 return string;
	   }

	   new = (char*) malloc ((*len)+2);
	   memcpy (new, string, *len);
	   new[(*len)] = '/';
	   new[(*len)+1] = '\0';
	   free (string);

	   (*len)++;

	   return new;
}


cherokee_plugin_table_entry_t *
plugin_table_entry_new (void)
{
	   cherokee_plugin_table_entry_t *entry;

	   cherokee_plugin_table_entry_new (&entry);
	   current_plugin_table_entry = entry;

	   return entry;
}


void
yyerror(char* msg)
{
        PRINT_ERROR ("Error parsing configuration: '%s', line %d, symbol '%s'\n", 
				 msg, yylineno, yytext);
}

%}



%token T_QUOTE
%token T_TIMEOUT T_KEEPALIVE T_DOCUMENT_ROOT T_LOG T_MIME_FILE T_DIRECTORY T_HANDLER T_USER T_GROUP 
%token T_SERVER T_USERDIR T_URL T_PIDFILE T_LISTEN T_FILEDIR T_HIDE_VERSION T_ENCODER T_ALLOW T_DENY

%token <number> T_NUMBER T_PORT 
%token <string> T_QSTRING T_FULLDIR T_ID T_HTTP_URL T_HOSTNAME T_IP 

%type <name_ptr> directory_option handler
%type <string> host_name
%type <list> id_list

%%

conffile : /* Empty */
         | lines
         ; 

lines : line
	 | lines line
	 ;

server_lines : server_line
             | server_lines server_line
             ;

line : server_line
     | common_line
     ;

common_line : server
            | port
            | listen
            | log
            | hideversion
            | mime
            | timeout
            | keepalive
            | pidfile
            | user1 
            | user2
            | group1 
            | group2
            | userdir
            ;

server_line : directory
            | document_root
            | encoder
            ;

handler_options : 
                | handler_options handler_option
                ;

directory_options : 
                | directory_options directory_option
                ;

encoder_options :
                | encoder_options encoder_option
                ;

id_list : T_ID 
{
	   linked_list_t *n = (linked_list_t *) malloc(sizeof(linked_list_t));
	   n->next = NULL;
	   n->string = $1;

	   $$ = n;
};

id_list : T_ID ',' id_list
{
	   linked_list_t *n = (linked_list_t *) malloc(sizeof(linked_list_t));

	   n->next = $3;
	   n->string = $1;

	   $$ = n;
};


port : T_PORT T_NUMBER 
{
	   SRV(server)->port = $2;
};

listen : T_LISTEN host_name
{
	   SRV(server)->listen_to = $2;
};

document_root : T_DOCUMENT_ROOT T_FULLDIR
{
	   cherokee_virtual_server_t *vserver;

	   vserver = auto_virtual_server;
	   vserver->root = $2;
	   vserver->root_len = strlen ($2);

	   /* Check for the endding slash
	    */
	   vserver->root = make_finish_with_slash (vserver->root, &(vserver->root_len));
};


log : T_LOG T_ID
{
	   /* Load the library
	    */
	   cherokee_logger_table_load (SRV(server)->loggers, $2);
}
log_optional
{
	   cherokee_virtual_server_t *vserver;
	   vserver = auto_virtual_server;

	   /* Instance the logger object
	    */
	   cherokee_logger_table_new_logger (SRV(server)->loggers, $2, vserver->logger_props, &vserver->logger);
};

log_optional : 
             | '{' tuple_list '}';

tuple_list : 
           | tuple_list tuple
           ;

tuple : T_ID T_FULLDIR
{
	   cherokee_virtual_server_t *vserver;
	   vserver = auto_virtual_server;

	   if (vserver->logger_props == NULL) {
			 cherokee_table_new (&vserver->logger_props);
	   }

	   cherokee_table_add (vserver->logger_props, $1, $2);
};


hideversion : T_HIDE_VERSION T_NUMBER
{
	   SRV(server)->hideversion = $2;
};

mime : T_MIME_FILE T_FULLDIR
{
	   cherokee_mime_t *mime;

	   cherokee_mime_init_with_file (&mime, (char *)$2);
	   free ($2);
};

timeout : T_TIMEOUT T_NUMBER
{
	   SRV(server)->timeout = $2;
};

keepalive : T_KEEPALIVE T_NUMBER
{
	   SRV(server)->keepalive = $2;
};

encoder : T_ENCODER T_ID
{
	   cherokee_encoder_table_load (SRV(server)->encoders, $2);
};

encoder : T_ENCODER T_ID '{'
{
	   cherokee_encoder_table_entry_t *enc;

	   cherokee_encoder_table_load (SRV(server)->encoders, $2);
	   cherokee_encoder_table_get (SRV(server)->encoders, $2, &enc);

	   current_encoder_entry = enc;
} 
encoder_options '}' 
{
	   current_encoder_entry = NULL;
};

encoder_option : T_ALLOW id_list
{
	   linked_list_t *i;
	   cherokee_matching_list_t *matching;

	   cherokee_matching_list_new (&matching);
	   cherokee_encoder_entry_set_matching_list (current_encoder_entry, matching);

	   i = $2;
	   while (i!=NULL) {
			 linked_list_t *prev;

			 cherokee_matching_list_add_allow (matching, i->string);

			 free(i->string);
			 prev = i;
			 i = i->next;
			 free(prev);
	   }

};

encoder_option : T_DENY id_list
{
	   linked_list_t *i;
	   cherokee_matching_list_t *matching;

	   cherokee_matching_list_new (&matching);
	   cherokee_encoder_entry_set_matching_list (current_encoder_entry, matching);

	   i = $2;
	   while (i!=NULL) {
			 linked_list_t *prev;

			 cherokee_matching_list_add_deny (matching, i->string);

			 free(i->string);
			 prev = i;
			 i = i->next;
			 free(prev);
	   }
};

pidfile : T_PIDFILE T_FULLDIR
{
	   FILE *file;

	   file = fopen ($2, "w");
	   if (file == NULL) {
			 PRINT_ERROR ("Can't write PID file '%s'\n", $2);
			 return 1;
	   }

	   snprintf (gbl_buffer, gbl_buffer_size, "%d\n", getpid());
	   fwrite (gbl_buffer, 1, strlen(gbl_buffer), file);
	   fclose (file);

	   free ($2);
};

user1 : T_USER T_ID
{
	   struct passwd *pwd;
	   
	   pwd = (struct passwd *) getpwnam ($2);
	   if (pwd == NULL) {
			 PRINT_ERROR ("User '%s' not found in the system", $2);
			 return 1;
	   }

	   SRV(server)->user = pwd->pw_uid;

	   free ($2);
};

user2 : T_USER T_NUMBER
{
	   SRV(server)->user = $2;
};

group1 : T_GROUP T_ID
{
	   struct group *grp;

	   grp = (struct group *) getgrnam ($2);
	   if (grp == NULL) {
			 PRINT_ERROR ("Group '%s' not found in the system", $2);
			 return 1;
	   }

	   SRV(server)->group = grp->gr_gid;

	   free ($2);
};

group2 : T_GROUP T_NUMBER
{
	   SRV(server)->group = $2;
};

handler : T_HANDLER T_ID '{' handler_options '}' {
	   $$.name = $2;
	   $$.ptr = current_plugin_table_entry;
};

handler : T_HANDLER T_ID 
{
	   $$.name = $2;
	   $$.ptr = current_plugin_table_entry;
};


handler_option : T_URL T_HTTP_URL
{
	   cherokee_plugin_table_entry_set (current_plugin_table_entry, "url", $2);
};

handler_option : T_FILEDIR T_FULLDIR
{
	   cherokee_plugin_table_entry_set (current_plugin_table_entry, "filedir", $2);
};


host_name : T_ID
          | T_IP 
{
	   $$ = $1;
};

server : T_SERVER host_name '{' 
{
	   CHEROKEE_NEW(vsrv, virtual_server);
	   current_virtual_server = vsrv;

	   vsrv->domain = $2;
	   cherokee_table_add (SRV(server)->vservers, $2, vsrv);

} server_lines '}' {

	   current_virtual_server = NULL;
};

directory : T_DIRECTORY T_FULLDIR '{' 
{
	   int name_len;

	   name_len = strlen($2);

	   /* Fill the tmp struct
	    */
	   directory_content_tmp.vserver        = auto_virtual_server;
	   directory_content_tmp.entry          = plugin_table_entry_new ();       /* new! */
	   directory_content_tmp.directory_name = make_finish_with_slash ($2, &name_len);
	   directory_content_tmp.handler_name   = NULL;
	   directory_content_tmp.handler        = NULL;
	   directory_content_tmp.document_root  = NULL;
} 
directory_options '}'
{
	   /* Basic checks
	    */
	   if (directory_content_tmp.handler_name == NULL) {
			 PRINT_ERROR ("Directory %s needs a handler; this directory entry is ignored.\n", directory_content_tmp.directory_name);
			 goto out;
	   }


	   /* Add the directory info
	    */
	   cherokee_plugin_table_fill_entry (directory_content_tmp.vserver->plugins, 
								  directory_content_tmp.handler_name, 
								  directory_content_tmp.entry);

	   if (directory_content_tmp.document_root != NULL) {

			 cherokee_buffer_make_empty(directory_content_tmp.entry->document_root);
			 cherokee_buffer_add (directory_content_tmp.entry->document_root,
							  directory_content_tmp.document_root,
							  strlen (directory_content_tmp.document_root));
	   }

 	   cherokee_virtual_server_set (directory_content_tmp.vserver, 
							  directory_content_tmp.directory_name, 
							  directory_content_tmp.handler_name,
							  directory_content_tmp.entry);

	   /* Clean
	    */
out:
	   if (directory_content_tmp.document_root != NULL) {
			 free (directory_content_tmp.document_root);
			 directory_content_tmp.document_root = NULL;
	   }
	   directory_content_tmp.vserver       = NULL;
	   directory_content_tmp.entry         = NULL;
	   directory_content_tmp.handler_name  = NULL;
	   directory_content_tmp.handler       = NULL;

	   current_plugin_table_entry = NULL;
};


directory_option : handler
{	   
	   directory_content_tmp.handler_name = $1.name;
	   directory_content_tmp.handler = $1.ptr;
};

directory_option : T_DOCUMENT_ROOT T_FULLDIR
{
	   int len;
	   char *new_fdir;

	   len = strlen($2);
	   new_fdir = make_finish_with_slash ($2, &len);

	   directory_content_tmp.document_root = new_fdir;
};

userdir : T_USERDIR T_ID '{' handler '}'
{
	   int len;
	   cherokee_plugin_table_entry_t *dir;

	   /* Have to finish with a slash
	    */
	   len = strlen($2);
	   SRV(server)->userdir = make_finish_with_slash ($2, &len);

	   dir = plugin_table_entry_new ();
	   cherokee_plugin_table_fill_entry (SRV(server)->vserver_default->plugins, $4.name, dir);
	   SRV(server)->userdir_handler = dir;

	   free ($4.name);
};


%%

